@@ -4,8 +4,9 @@ import logging
 import six
 
 import numpy as np
+import scipy as sp
 
-from kalepy import kernels, utils
+from kalepy import kernels, utils, _NUM_PAD
 from kalepy import _BANDWIDTH_DEFAULT
 
 __all__ = ['KDE']
@@ -164,7 +165,7 @@ class KDE(object):
 
         """
         self._helper = helper
-        self.dataset = np.atleast_2d(dataset)
+        self._dataset = np.atleast_2d(dataset)
         ndim, ndata = self.dataset.shape
         if weights is None:
             weights = np.ones(ndata)/ndata
@@ -198,19 +199,74 @@ class KDE(object):
 
         self._neff = neff
 
-        data_cov = np.cov(dataset, rowvar=True, bias=False, aweights=weights)
-        self._data_cov = np.atleast_2d(data_cov)
-        self.set_bandwidth(bandwidth, bw_rescale)
+        covariance = np.cov(dataset, rowvar=True, bias=False, aweights=weights)
+        self._covariance = np.atleast_2d(covariance)
+        self._set_bandwidth(bandwidth, bw_rescale)
 
         # Convert from string, class, etc to a kernel
         dist = kernels.get_distribution_class(kernel)
         self._kernel = kernels.Kernel(
-            distribution=dist, matrix=self.matrix, helper=helper, **kwargs)
+            distribution=dist, bandwidth=self._bandwidth, covariance=self._covariance,
+            helper=helper, **kwargs)
 
         self._finalize()
+        self._cdf_grid = None
+        self._cdf_func = None
         return
 
-    def pdf(self, pnts, **kwargs):
+    def _guess_edges(self, nmin=100, nmax=1e4, tot_max=1e8, reflect=None):
+        """Guess good grid edges to resolve the KDE's PDF.
+
+        To-Do: this doesn't need to be done on a grid, do something smarter!
+        To-Do: implement reflect (set boundaries to reflect values when given)
+
+        """
+        data = self.dataset
+        neff = self.neff
+        ndim = self.ndim
+        bandwidth = self.kernel.bandwidth.diagonal()
+        finite = self.kernel.FINITE
+        if reflect is not None:
+            raise ValueError("`reflect` is not supported in calculing grid edges!")
+
+        # Number of grid points in each dimension
+        max_per_dim = np.power(tot_max, 1.0/ndim)
+        if max_per_dim < nmin:
+            err = (
+                "Cannot resolve ndim={} dimensions ".format(ndim) +
+                "with at least nmin={:.2e} ".format(nmin) +
+                "points and stay below maximum = {:.3e}!".format(tot_max)
+            )
+            raise ValueError(err)
+
+        npd = np.power(neff, 1.0/ndim)
+        npd = np.clip(npd, nmin, nmax)
+        npd = np.clip(npd, 0.0, max_per_dim)
+        npd = int(npd)
+        msg = "KDE._guess_edges:: Using {} bins per dimension".format(npd)
+        # print(msg)
+        if self._helper:
+            logging.info(msg)
+
+        # If the Kernel is finite, then there is only support out to `bandwidth` beyond extrema
+        if finite:
+            out = (1.0 + _NUM_PAD)
+        # If infinite, how many standard-deviations can we expect data points to lie at
+        else:
+            out = sp.stats.norm.ppf(1 - 1/neff)
+            # Extra to be double sure...
+            out *= 1.5
+
+        # Find the extrema in each dimension
+        extr = [[np.min(dd) - bw*out, np.max(dd) + bw*out] for bw, dd in zip(bandwidth, data)]
+        edges = [np.linspace(*ex, npd) for ex in extr]
+        msg = "KDE._guess_edges:: extrema = {}".format(
+            ", ".join(["[{:.2e}, {:.2e}]".format(*ex) for ex in extr]))
+        # print(msg)
+
+        return edges
+
+    def pdf(self, pnts, reflect=None, params=None):
         """Evaluate the kernel-density-estimate PDF at the given data-points.
 
         This method acts as an API to the `Kernel.pdf` method of this instance's `kernel`.
@@ -223,24 +279,21 @@ class KDE(object):
             NOTE: If the `params` kwarg (see below) is given, then only those dimensions of the
             target parameters should be specified in `pnts`.
 
-        kwargs ::
-            Additional, optional keyword arguments passed to `Kernel.pdf`.  Accepted arguments:
-
-            reflect : (D,) array_like, None (default)
-                Locations at which reflecting boundary conditions should be imposed.
-                For each dimension `D`, a pair of boundary locations (for: lower, upper) must be
-                specified, or `None`.  `None` can also be given to specify no boundary at that
-                location.  See class docstrings:`Reflection` for more information.
-
-            params : int, array_like of int, None (default)
-                Only calculate the PDF for certain parameters (dimensions).
-                See class docstrings:`Projection` for more information.
+        reflect : (D,) array_like, None (default)
+            Locations at which reflecting boundary conditions should be imposed.
+            For each dimension `D`, a pair of boundary locations (for: lower, upper) must be
+            specified, or `None`.  `None` can also be given to specify no boundary at that
+            location.  See class docstrings:`Reflection` for more information.
+        params : int, array_like of int, None (default)
+            Only calculate the PDF for certain parameters (dimensions).
+            See class docstrings:`Projection` for more information.
 
         """
-        result = self.kernel.pdf(pnts, self.dataset, self.weights, **kwargs)
+        result = self.kernel.pdf(pnts, self.dataset, self.weights, reflect=reflect, params=params)
+        # print("KDE.pdf(): result.shape = {}".format(result.shape))
         return result
 
-    def pdf_grid(self, edges, **kwargs):
+    def pdf_grid(self, edges, reflect=None, params=None):
         """Convenience method to compute the PDF given the edges of a grid in each dimension.
 
         Arguments
@@ -255,20 +308,78 @@ class KDE(object):
             parameter.  Define the lengths of each array as: A, B, C; then the grid and the
             returned PDF will have a shape (A, B, C).
 
-        kwargs ::
-            Additional, optional keyword arguments passed to `Kernel.pdf`.  Accepted arguments:
+        reflect : (D,) array_like, None (default)
+            Locations at which reflecting boundary conditions should be imposed.
+            For each dimension `D`, a pair of boundary locations (for: lower, upper) must be
+            specified, or `None`.  `None` can also be given to specify no boundary at that
+            location.  See class docstrings:`Reflection` for more information.
 
-            reflect : (D,) array_like, None (default)
-                Locations at which reflecting boundary conditions should be imposed.
-                For each dimension `D`, a pair of boundary locations (for: lower, upper) must be
-                specified, or `None`.  `None` can also be given to specify no boundary at that
-                location.  See class docstrings:`Reflection` for more information.
-
-            params : int, array_like of int, None (default)
-                Only calculate the PDF for certain parameters (dimensions).
-                See class docstrings:`Projection` for more information.
+        params : int, array_like of int, None (default)
+            Only calculate the PDF for certain parameters (dimensions).
+            See class docstrings:`Projection` for more information.
 
         """
+        ndim = self.ndim
+        # If `edges` is 1D (and not "jagged") then expand to 2D
+        if utils.really1d(edges):
+            edges = np.atleast_2d(edges)
+
+        # Check shape if unput `edges`
+        if (params is None):
+            # Should be edges for all parameters
+            if len(edges) != ndim:
+                err = "`edges` must be (D,)=({},): an arraylike of edges for each dim/param!"
+                err = err.format(ndim)
+                raise ValueError(err)
+        else:
+            # Should only be as many edges as target parameters
+            if np.size(params) != len(edges):
+                err = "`params` = '{}' (length {}), but length of `edges` = {}!".format(
+                    params, len(params), len(edges))
+                raise ValueError(err)
+
+        coords = np.meshgrid(*edges, indexing='ij')
+        shp = np.shape(coords)[1:]
+        coords = np.vstack([xx.ravel() for xx in coords])
+        pdf = self.pdf(coords, reflect=reflect, params=params)
+        pdf = pdf.reshape(shp)
+        # print("KDE.pdf_grid(): result.shape = {}".format(pdf.shape))
+        return pdf
+
+    def cdf(self, pnts):
+        """Cumulative Distribution Function based on KDE smoothed data.
+
+        Arguments
+        ---------
+        pnts : ([D,]N,) array_like of scalar
+            Target evaluation points
+
+        Returns
+        -------
+        cdf : (N,) ndarray of scalar
+            CDF Values at the target points
+
+        """
+        if self._cdf_func is None:
+            edges = self._guess_edges()
+
+            # Calculate PDF at grid locations
+            pdf = self.pdf_grid(edges)
+            # Convert to CDF using trapezoid rule
+            cdf = utils.cumtrapz(pdf, edges)
+            # Normalize to the maximum value
+            cdf /= cdf.max()
+
+            self._cdf_grid = (edges, cdf)
+            self._cdf_func = sp.interpolate.RegularGridInterpolator(
+                *self._cdf_grid, bounds_error=False, fill_value=None)
+
+        # `scipy.interplate.RegularGridInterpolator` expects shape (N,D,) -- so transpose
+        pnts = np.asarray(pnts).T
+        cdf = self._cdf_func(pnts)
+        return cdf
+
+    def cdf_grid(self, edges, **kwargs):
         ndim = self.ndim
         if len(edges) != ndim:
             err = "`edges` must be (D,)=({},): an arraylike of edges for each dim/param!"
@@ -278,9 +389,9 @@ class KDE(object):
         coords = np.meshgrid(*edges, indexing='ij')
         shp = np.shape(coords)[1:]
         coords = np.vstack([xx.ravel() for xx in coords])
-        pdf = self.pdf(coords, **kwargs)
-        pdf = pdf.reshape(shp)
-        return pdf
+        cdf = self.cdf(coords, **kwargs)
+        cdf = cdf.reshape(shp)
+        return cdf
 
     def resample(self, size=None, keep=None, reflect=None, squeeze=True):
         """Draw new values from the kernel-density-estimate calculated PDF.
@@ -321,10 +432,15 @@ class KDE(object):
             size=size, keep=keep, reflect=reflect, squeeze=squeeze)
         return samples
 
-    def _finalize(self, log_ratio_tol=5.0):
+    # def ppf(self, cfrac):
+    #     return self.kernel.ppf(cfrac)
 
-        mat = self.matrix
-        ndim = self.ndim
+    def _finalize(self, log_ratio_tol=5.0):
+        WARN_ALL_BELOW = -10
+        EXTR_ABOVE = -20
+
+        mat = self.kernel.matrix
+        ndim = self._ndim
         # Diagonal elements of the matrix
         diag = mat.diagonal()
 
@@ -333,100 +449,126 @@ class KDE(object):
         li = np.tril_indices(ndim, -1)
         offd = np.append(mat[ui], mat[li])
 
+        warn = False
+
         if np.any(offd != 0.0):
-            d_extr = utils.minmax(np.log10(np.fabs(diag)))
-            o_extr = utils.minmax(np.log10(np.fabs(offd)))
+            d_vals = np.log10(np.fabs(diag))
+            o_vals = np.log10(np.fabs(offd))
+            if np.all(d_vals <= WARN_ALL_BELOW) and np.all(o_vals <= WARN_ALL_BELOW):
+                logging.warning("Covariance matrix:\n" + str(mat))
+                logging.warning("All matrix elements are less than 1e{}!".format(WARN_ALL_BELOW))
+                warn = True
+
+            d_extr = utils.minmax(d_vals[d_vals > EXTR_ABOVE])
+            o_extr = utils.minmax(o_vals[o_vals > EXTR_ABOVE])
+
             ratio = np.fabs(d_extr[:, np.newaxis] - o_extr[np.newaxis, :])
             if np.any(ratio >= log_ratio_tol):
                 logging.warning("Covariance matrix:\n" + str(mat))
                 msg = "(log) Ratio of covariance elements ({}) exceeds tolerance ({})!".format(
                     ratio, log_ratio_tol)
                 logging.warning(msg)
-                msg = "Recommend rescaling input data, or using a diagonal covariance matrix"
-                logging.warning(msg)
+                warn = True
+
+        if warn:
+            msg = "Recommend rescaling input data, or using a diagonal covariance matrix"
+            logging.warning(msg)
 
         return
 
     # ==== Properties ====
 
     @property
-    def weights(self):
-        return self._weights
+    def bandwidth(self):
+        return self._bandwidth
 
     @property
-    def neff(self):
-        return self._neff
+    def covariance(self):
+        return self._covariance
 
     @property
-    def ndim(self):
-        return self._ndim
+    def dataset(self):
+        return self._dataset
+
+    @property
+    def kernel(self):
+        return self._kernel
 
     @property
     def ndata(self):
         return self._ndata
 
     @property
-    def kernel(self):
-        return self._kernel
+    def ndim(self):
+        return self._ndim
+
+    @property
+    def neff(self):
+        return self._neff
+
+    @property
+    def weights(self):
+        return self._weights
 
     # ==== BANDWIDTH ====
 
-    def set_bandwidth(self, bandwidth, bw_rescale):
+    def _set_bandwidth(self, bw_input, bw_rescale):
         ndim = self.ndim
-        _input = bandwidth
-        bw_white_matrix = np.zeros((ndim, ndim))
+        bandwidth = np.zeros((ndim, ndim))
 
-        if len(np.atleast_1d(bandwidth)) == 1:
-            _bw, method = self._compute_bandwidth(bandwidth)
+        if len(np.atleast_1d(bw_input)) == 1:
+            _bw, method = self._compute_bandwidth(bw_input)
             if not self._diagonal:
-                bw_white_matrix[...] = _bw
+                bandwidth[...] = _bw
             else:
                 idx = np.arange(ndim)
-                bw_white_matrix[idx, idx] = _bw
+                bandwidth[idx, idx] = _bw
         else:
-            if np.shape(bandwidth) == (ndim,):
+            if np.shape(bw_input) == (ndim,):
                 # bw_method = 'diagonal'
                 for ii in range(ndim):
-                    bw_white_matrix[ii, ii] = self._compute_bandwidth(
-                        bandwidth[ii], param=(ii, ii))[0]
+                    bandwidth[ii, ii], *_ = self._compute_bandwidth(
+                        bw_input[ii], param=(ii, ii))
                 method = 'diagonal'
-            elif np.shape(bandwidth) == (ndim, ndim):
+            elif np.shape(bw_input) == (ndim, ndim):
                 for ii, jj in np.ndindex(ndim, ndim):
-                    bw_white_matrix[ii, jj] = self._compute_bandwidth(
-                        bandwidth[ii, jj], param=(ii, jj))[0]
+                    bandwidth[ii, jj], *_ = self._compute_bandwidth(
+                        bw_input[ii, jj], param=(ii, jj))
                 method = 'matrix'
             else:
-                raise ValueError("`bandwidth` have shape (1,), (N,) or (N,) for `N` dimensions!")
+                err = "`bandwidth` must have shape (1,), (N,) or (N,N,) for `N` dimensions!"
+                raise ValueError(err)
 
-        if self._helper and np.any(np.isclose(bw_white_matrix.diagonal(), 0.0)):
-            ii = np.where(np.isclose(bw_white_matrix.diagonal(), 0.0))[0]
+        if self._helper and np.any(np.isclose(bandwidth.diagonal(), 0.0)):
+            ii = np.where(np.isclose(bandwidth.diagonal(), 0.0))[0]
             msg = "WARNING: diagonal '{}' of bandwidth is near zero!".format(ii)
             logging.warning(msg)
 
         # Rescale the bandwidth matrix
         if bw_rescale is not None:
             bwr = np.atleast_2d(bw_rescale)
-            bw_white_matrix = bwr * bw_white_matrix
+            bandwidth = bwr * bandwidth
             if self._helper:
                 logging.info("Rescaling `bw_white_matrix` by '{}'".format(
                     bwr.squeeze().flatten()))
 
-        matrix = self.data_cov * (bw_white_matrix ** 2)
+        # bw_matrix = self._data_cov * (bw_white_matrix ** 2)
+        self._bandwidth = bandwidth
 
         # prev: bw_white
-        self._bw_white_matrix = bw_white_matrix
+        # self._bw_white_matrix = bw_white_matrix
         self._method = method
-        self._input = _input
+        # self._bw_input = bw_input
         # prev: bw_cov
-        self._matrix = matrix
+        # self._bw_matrix = bw_matrix
         return
 
     def _compute_bandwidth(self, bandwidth, param=None):
         if isinstance(bandwidth, six.string_types):
             if bandwidth == 'scott':
-                bw = self.scott_factor(param=param)
+                bw = self._scott_factor(param=param)
             elif bandwidth == 'silverman':
-                bw = self.silverman_factor(param=param)
+                bw = self._silverman_factor(param=param)
             else:
                 msg = "Unrecognized bandwidth str specification '{}'!".format(bandwidth)
                 raise ValueError(msg)
@@ -437,25 +579,20 @@ class KDE(object):
             bw = bandwidth
             method = 'constant scalar'
 
-        elif callable(bandwidth):
-            bw = bandwidth(self, param=param)
-            method = 'function'
-
         else:
             raise ValueError("Unrecognized `bandwidth` '{}'!".format(bandwidth))
 
+        '''
+        elif callable(bandwidth):
+            bw = bandwidth(self, param=param)
+            method = 'function'
+            bw_cov = 1.0
+        '''
+
         return bw, method
 
-    def scott_factor(self, *args, **kwargs):
+    def _scott_factor(self, *args, **kwargs):
         return np.power(self.neff, -1./(self.ndim+4))
 
-    def silverman_factor(self, *args, **kwargs):
+    def _silverman_factor(self, *args, **kwargs):
         return np.power(self.neff*(self.ndim+2.0)/4.0, -1./(self.ndim+4))
-
-    @property
-    def matrix(self):
-        return self._matrix
-
-    @property
-    def data_cov(self):
-        return self._data_cov
