@@ -7,14 +7,16 @@ import re
 import numpy as np
 import scipy as sp
 import scipy.linalg  # noqa
-import matplotlib.pyplot as plt
 
+
+'''
 __all__ = [
-    'add_cov', 'array_str', 'ave_std', 'bins', 'check_path', 'cumtrapz', 'midpoints',
-    'minmax', 'modify_exists', 'nbshow', 'percentiles', 'rem_cov',
-    'save_fig', 'spacing', 'stats_str',
+    'add_cov', 'array_str', 'ave_std', 'bins', 'check_path', 'cumsum', 'cumtrapz', 'midpoints',
+    'minmax', 'modify_exists', 'percentiles', 'rem_cov',
+    'spacing', 'stats_str',
     'trapz_nd', 'trapz_dens_to_mass'
 ]
+'''
 
 
 def add_cov(data, cov):
@@ -46,6 +48,17 @@ def array_str(data, num=3, format=':.2e'):
     return rv
 
 
+def assert_true(val, msg=None):
+    msg_succ, msg_fail = _prep_msg(msg)
+    if not val:
+        raise AssertionError(msg_fail)
+
+    if msg_succ is not None:
+        print(msg_succ)
+
+    return
+
+
 def allclose(xx, yy, msg=None, **kwargs):
     msg_succ, msg_fail = _prep_msg(msg)
     xx = np.atleast_1d(xx)
@@ -93,8 +106,17 @@ def ave_std(values, weights=None, **kwargs):
     return average, np.sqrt(variance)
 
 
-def bins(*args):
-    xe = np.linspace(*args)
+def bins(*args, **kwargs):
+    """Calculate `np.linspace(*args)` and return also centers and widths.
+
+    Returns
+    -------
+    xe : (N,) bin edges
+    xc : (N-1,) bin centers
+    dx : (N-1,) bin widths
+
+    """
+    xe = np.linspace(*args, **kwargs)
     xc = midpoints(xe)
     dx = np.diff(xe)
     return xe, xc, dx
@@ -191,6 +213,74 @@ def cov_from_var_cor(var, corr):
     return cov
 
 
+def cumsum(vals, axis=None):
+    """Perform a cumulative sum without flattening the input array.
+
+    See: https://stackoverflow.com/a/60647166/230468
+
+    Arguments
+    ---------
+    vals : array_like of scalar
+        Input values to sum over.
+    axis : None or int
+        Axis over which to perform the cumulative sum.
+
+    Returns
+    -------
+    res : ndarray of scalar
+        Same shape as input `vals`
+
+    """
+
+    vals = np.asarray(vals)
+    nd = np.ndim(vals)
+    if (axis is not None) or (nd == 1):
+        return np.cumsum(vals, axis=axis)
+
+    res = vals.cumsum(-1)
+    for ii in range(2, nd+1):
+        np.cumsum(res, axis=-ii, out=res)
+
+    return res
+
+
+def cumtrapz(pdf, edges, prepend=True, axis=None):
+    """Perform a cumulative integration using the trapezoid rule.
+
+    Arguments
+    ---------
+    pdf : array_like of scalar
+        Input values (e.g. a PDF) to be integrated.
+    edges : [D,] list of (array_like of scalar)
+        Edges defining bins along each dimension.
+        This should be an array/list of edges for each of `D` dimensions.
+    prepend : bool
+        Whether or not to prepend zero values along the integrated dimensions.
+    axis : None or int
+        Axis/Dimension over which to integrate.
+
+    Returns
+    -------
+    cdf : ndarray of scalar
+        Values integrated over the desired axes.
+        Shape:
+            If `prepend` is False, the shape of `cdf` will be one smaller than the input `pdf`
+            in all dimensions integrated over.
+            If `prepend` is True, the shape of `cdf` will match that of the input `pdf`.
+
+    """
+    # Convert from density to mass using trapezoid rule in each bin
+    pmf = trapz_dens_to_mass(pdf, edges, axis=axis)
+    # Perform cumulative sumation
+    cdf = cumsum(pmf, axis=axis)
+
+    # Prepend zeros to output array
+    if prepend:
+        cdf = _pre_pad_zero(cdf, axis=axis)
+
+    return cdf
+
+
 def matrix_invert(matrix, helper=True):
     try:
         matrix_inv = np.linalg.inv(matrix)
@@ -310,10 +400,6 @@ def modify_exists(path_fname):
     return fname
 
 
-def nbshow():
-    return run_if_notebook(plt.show)
-
-
 def percentiles(values, percs=None, sigmas=None, weights=None, axis=None, values_sorted=False):
     """Compute weighted percentiles.
 
@@ -370,6 +456,33 @@ def percentiles(values, percs=None, sigmas=None, weights=None, axis=None, values
     return percs
 
 
+def really1d(arr):
+    """Test whether an array_like is really 1D (i.e. not a jagged ND array).
+
+    Test whether the input array is uniformly one-dimensional, as apposed to (e.g.) a ``ndim == 1``
+    list or array of irregularly shaped sub-lists/sub-arrays.  True for an empty list `[]`.
+
+    Arguments
+    ---------
+    arr : array_like
+        Array to be tested.
+
+    Returns
+    -------
+    bool
+        Whether `arr` is purely 1D.
+
+    """
+    if np.ndim(arr) != 1:
+        return False
+    # Empty list or array
+    if len(arr) == 0:
+        return True
+    if np.any(np.vectorize(np.ndim)(arr)):
+        return False
+    return True
+
+
 def rem_cov(data, cov=None):
     if cov is None:
         cov = np.cov(*data)
@@ -379,21 +492,6 @@ def rem_cov(data, cov=None):
     return white_data
 
 
-def save_fig(fig, fname, path=None, quiet=False, rename=True, **kwargs):
-    """Save the given figure to the given filename, with some added niceties.
-    """
-    if path is None:
-        path = os.path.abspath(os.path.curdir)
-    fname = os.path.join(path, fname)
-    check_path(fname)
-    if rename:
-        fname = modify_exists(fname)
-    fig.savefig(fname, **kwargs)
-    if not quiet:
-        print("Saved to '{}'".format(fname))
-    return fname
-
-
 def spacing(data, scale='log', num=None, dex=10, **kwargs):
     DEF_NUM_LIN = 20
 
@@ -414,6 +512,7 @@ def spacing(data, scale='log', num=None, dex=10, **kwargs):
         else:
             num = DEF_NUM_LIN
 
+    num = int(num)
     if log_flag:
         spaced = np.logspace(*np.log10(span), num=num)
     else:
@@ -519,12 +618,6 @@ def trapz_nd(data, edges, axis=None):
     return tot
 
 
-def cumtrapz(pdf, edges, **kwargs):
-    pmf = trapz_dens_to_mass(pdf, edges, **kwargs)
-    cdf = np.cumsum(pmf)
-    return cdf
-
-
 def trapz_dens_to_mass(pdf, edges, axis=None):
     """Convert from density to mass, for values on the corner of a grid, using the trapezoid rule.
 
@@ -622,22 +715,24 @@ def trapz_dens_to_mass(pdf, edges, axis=None):
     return mass
 
 
-def run_if(func, target, *args, **kwargs):
+def run_if(func, target, *args, otherwise=None, **kwargs):
     env = _python_environment()
     if env.startswith(target):
         return func(*args, **kwargs)
+    elif otherwise is not None:
+        return otherwise(*args, **kwargs)
 
     return None
 
 
-def run_if_notebook(func, *args, **kwargs):
+def run_if_notebook(func, *args, otherwise=None, **kwargs):
     target = 'notebook'
-    return run_if(func, target, *args, **kwargs)
+    return run_if(func, target, *args, otherwise=otherwise, **kwargs)
 
 
-def run_if_script(func, *args, **kwargs):
+def run_if_script(func, *args, otherwise=None, **kwargs):
     target = 'script'
-    return run_if(func, target, *args, **kwargs)
+    return run_if(func, target, *args, otherwise=otherwise, **kwargs)
 
 
 def _is_notebook():
@@ -698,6 +793,16 @@ def _guess_str_format_from_range(arr, prec=2, log_limit=2, allow_int=True):
     return form
 
 
+def _pre_pad_zero(aa, axis=None):
+    if axis is None:
+        return np.pad(aa, [1, 0])
+
+    aa = np.moveaxis(aa, axis, 0)
+    aa = np.concatenate([[np.zeros_like(aa[0])], aa], axis=0)
+    aa = np.moveaxis(aa, 0, axis)
+    return aa
+
+
 def _prep_msg(msg=None):
     if (msg is None) or (msg is True):
         msg_fail = "FAILURE:: arrays do not match!"
@@ -751,31 +856,3 @@ class Test_Base(object):
 
         np.random.seed(object.__getattribute__(self, "DEF_SEED"))
         return value
-
-
-class plot_control:
-
-    def __init__(self, fname, *args, **kwargs):
-        self.fname = fname
-        self.args = args
-        self.kwargs = kwargs
-        return
-
-    def __enter__(self):
-        # if not _is_notebook():
-        #     raise _DummyError
-
-        plt.close('all')
-        return self
-
-    def __exit__(self, type, value, traceback):
-        # if isinstance(value, _DummyError):
-        #     return True
-
-        plt.savefig(self.fname, *self.args, **self.kwargs)
-        if _is_notebook():
-            plt.show()
-        else:
-            plt.close('all')
-
-        return
